from easy_ecs_sim.system import System, SystemBag


class Systems:
    def __init__(self, systems: list[System] = None):
        self.systems = systems or []
        self.paused = set()

    def flatten(self):
        systems = []
        for _ in self.systems:
            if isinstance(_, SystemBag):
                systems.extend(_.steps)
            else:
                systems.append(_)
        return systems

    def actives(self):
        return [_ for _ in self.flatten() if _ not in self.paused]

    def insert(self, index: int, item: System):
        self.systems.insert(index, item)

    def append(self, item: System):
        self.systems.append(item)



from typing import Type, TypeVar

from easy_ecs_sim.context import Context

T = TypeVar('T')


class ContextService:
    def __init__(self, ctx: Context):
        self.ctx = ctx

    def find[T](self, ctype: Type[T]):
        return self.ctx.find(ctype)



import time
import traceback
from easy_kit.timing import time_func, timing

from easy_ecs_sim.component import Component
from easy_ecs_sim.context import Context
from easy_ecs_sim.storage.database import Database
from easy_ecs_sim.storage.demography import Demography
from easy_ecs_sim.storage.my_database import MyDatabase
from easy_ecs_sim.system import System
from easy_ecs_sim.systems import Systems
from easy_ecs_sim.types import EntityId
from easy_ecs_sim.utils import ComponentSet


class ECS:
    def __init__(self, ctx: Context = None):
        self.ctx: Context = ctx or Context()
        self.db = MyDatabase()
        self.systems = Systems()
        self.last_updates = {}
        self.is_running: bool = True

    def add_systems(self, *systems: System):
        for sys in systems:
            self.systems.append(sys)
        return self.connect(self.ctx)

    def connect(self, ctx: Context):
        self.ctx = ctx
        self.ctx.register(self)
        self.ctx.register(self.db, Database)
        for _ in self.systems.flatten():
            self.ctx.register(_)
        return self

    def set_running(self, status: bool = None):
        if status is None:
            status = not self.is_running
        self.is_running = status
        if self.is_running:
            self.last_updates = {}
        return self

    def create_all(self, items: list[ComponentSet]):
        self.db.create_all(items)
        return self

    @time_func
    def update(self):
        self.apply_demography()

        if not self.is_running:
            return

        now = time.time()
        for sys in self.systems.paused:
            self.last_updates[sys.__class__] = now

        for sys in self.systems.actives():
            sys_key = sys.__class__
            if sys_key not in self.last_updates:
                self.last_updates[sys_key] = now
            elapsed = now - self.last_updates[sys_key]
            if elapsed < sys.periodicity_sec:
                continue
            self.last_updates[sys_key] = now

            try:
                with timing(f'ECS.update.{sys.sys_id}'):
                    sys.update(self.ctx, self.db, elapsed)
            except Exception as e:
                print(f'{sys.sys_id}: {e}\n{traceback.format_exc()}')

    @time_func
    def apply_demography(self):
        status = Demography().load(self.db.dirty)
        self.db.dirty.clear()

        systems = [_ for _ in self.systems.flatten() if _._signature is not None]
        for sys in systems:
            for _ in status.death:
                self._handle_death(sys, _)
        for sys in systems:
            for _ in status.birth:
                self._handle_birth(sys, _)

        self.db.update_demography(status)

    def _handle_birth(self, sys: System, items: Component | list[Component]):
        if isinstance(items, Component):
            items = [items]
        for item in items:
            item.db = self.db
            item.ctx = self.ctx
        signature = sys._signature
        item = signature.cast(items)
        if item is not None:
            self.db.get_table(signature).create(item)
            sys.register(self.ctx, item)

    def _handle_death(self, sys: System, eid: EntityId):
        index = self.db.get_table(sys._signature)
        item = index.read(eid)
        if item:
            sys.unregister(self.ctx, item)
        index.destroy(eid)


# default simulator
sim = ECS()



from dataclasses import field
from typing import Self

import math
import numpy as np

from easy_ecs_sim.component import Component


class Vec3(Component):
    raw: np.ndarray = field(default_factory=lambda: np.zeros((3,)))

    @classmethod
    def create(cls, x: float = 0, y: float = 0, z: float = 0):
        return cls(raw=np.array([x, y, z], dtype=float))

    @staticmethod
    def direction(a: 'Vec3', b: 'Vec3'):
        return Vec3(raw=b.raw - a.raw)

    @property
    def x(self):
        return self.raw[0]

    @property
    def y(self):
        return self.raw[1]

    @property
    def z(self):
        return self.raw[2]

    @x.setter
    def x(self, value: float):
        self.raw[0] = value

    @y.setter
    def y(self, value: float):
        self.raw[1] = value

    @z.setter
    def z(self, value: float):
        self.raw[2] = value

    def norm(self):
        # return np.linalg.vector_norm(self.raw)
        return math.sqrt(self.x * self.x + self.y * self.y + self.z * self.z)

    def normal(self, size: float = 1.0):
        n = self.norm()
        if n == 0:
            return self
        return self * size / n

    @classmethod
    def random(cls, shape: Self = None):
        if shape is not None:
            return cls(raw=np.random.rand(3) * shape.raw)
        return cls(raw=np.random.rand(3))

    def __add__(self, other):
        if isinstance(other, Vec3):
            other = other.raw
        return self.__class__(raw=self.raw + other)

    def __iadd__(self, other):
        if isinstance(other, Vec3):
            other = other.raw
        self.raw += other
        return self

    def __sub__(self, other):
        if isinstance(other, Vec3):
            other = other.raw
        return self.__class__(raw=self.raw - other)

    def __isub__(self, other):
        if isinstance(other, Vec3):
            other = other.raw
        self.raw -= other
        return self

    def __mul__(self, other):
        if isinstance(other, Vec3):
            other = other.raw
        return self.__class__(raw=self.raw * other)

    def __imul__(self, other):
        if isinstance(other, Vec3):
            other = other.raw
        self.raw *= other
        return self

    def __truediv__(self, other):
        if isinstance(other, Vec3):
            other = other.raw
        return self.__class__(raw=self.raw / other)

    def __itruediv__(self, other):
        if isinstance(other, Vec3):
            other = other.raw
        self.raw /= other
        return self



import time
from dataclasses import dataclass, field
from typing import Type, TypeVar

from easy_ecs_sim.storage.database import Database

T = TypeVar('T')


@dataclass
class TimeboxManager[T]:
    signature: Type[T]
    max_time_per_step: float
    items: list[T] = field(default_factory=list)
    dt_bonus: float = 0

    def iter(self, db: Database, dt: float):
        if len(self.items) == 0:
            self.items = list(db.get_table(self.signature).iter())
            self.dt_bonus = 0

        start = time.time()
        while time.time() - start < self.max_time_per_step:
            if len(self.items) == 0:
                break
            yield self.items.pop()
        self.dt_bonus += dt



class IdGenerator:
    def __init__(self):
        self._last_id = -1

    def new_id(self):
        self._last_id += 1
        return self._last_id

    def gen(self, n: int):
        return [self.new_id() for _ in range(n)]



from dataclasses import field, dataclass
from typing import Self

from easy_kit.timing import time_func

from easy_ecs_sim.component import Component
from easy_ecs_sim.utils import ComponentSet, flatten_components
from easy_ecs_sim.signature import Signature
from easy_ecs_sim.types import EntityId


@dataclass
class Demography:
    birth: list[list[Component]] = field(default_factory=list)
    death: set[EntityId] = field(default_factory=set)

    def clear(self):
        self.birth.clear()
        self.death.clear()

    def with_birth(self, items: list[ComponentSet]):
        self.birth.extend(map(flatten_components, items))
        return self

    def with_death(self, items: Component | Signature | list[Component | Signature]):
        if isinstance(items, Component | Signature):
            items = [items]
        self.death.update([_.eid for _ in items])
        return self

    @time_func
    def load(self, other: Self):
        if other is None:
            return self
        self.death.update(other.death)
        self.birth.extend(other.birth)
        return self



from typing import Type, Any, TypeVar

T = TypeVar('T')


class Context:
    def __init__(self, *initial_state: Any):
        self.data: dict[Type, Any] = {}
        for _ in initial_state:
            self.register(_)

    def register[T](self, data: T, ctype: Type[T] = None):
        found_type = type(data)

        if ctype is not None:
            if not issubclass(found_type, ctype):
                raise ValueError(f'incompatible type [{ctype}] for data of type [{found_type}]')

        if ctype is not None:
            assert issubclass(found_type, ctype)
        else:
            ctype = found_type
        self.data[ctype] = data
        return self

    def find[T](self, ctype: Type[T]) -> T:
        return self.data[ctype]



from dataclasses import field, dataclass
from typing import Type, Iterable, Any, Generator, TypeVar

from easy_ecs_sim.component import Component
from easy_ecs_sim.signature import Signature
from easy_ecs_sim.types import EntityId

T = TypeVar('T')


@dataclass
class Index[T: Signature | Component]:
    ttype: Type[T]
    by_entity: dict[EntityId, T] = field(default_factory=dict)

    @property
    def entities(self):
        return set(self.by_entity.keys())

    def iter(self, eids: Iterable[EntityId] = None) -> Generator[T | None, Any, None]:
        if eids is None:
            yield from self.by_entity.values()
        else:
            for eid in eids:
                yield self.read(eid)

    @property
    def count(self):
        return len(self.by_entity)

    def create(self, item: T):
        self.by_entity[item.eid] = item

    def read(self, eid: EntityId):
        return self.by_entity.get(eid, None)

    def read_any(self):
        return next(iter(self.by_entity.values()))

    def destroy(self, eid: EntityId):
        return self.by_entity.pop(eid, None)

    def destroy_all(self, eids: Iterable[EntityId]):
        for _ in eids:
            self.destroy(_)

    def clear(self):
        self.by_entity.clear()



from dataclasses import field
from typing import Self, Type, Any, TypeVar

from easy_kit.my_model import MyModel
from easy_kit.timing import time_func

from easy_ecs_sim.context import Context
from easy_ecs_sim.storage.id_generator import IdGenerator
from easy_ecs_sim.types import ComponentId, EntityId

CID_GEN = IdGenerator()

T = TypeVar('T')


class Component(MyModel):
    cid: ComponentId = field(default_factory=CID_GEN.new_id)
    eid: EntityId = -1

    db: Any = None
    ctx: Context | None = None

    @classmethod
    def signature(cls) -> list[Type[Self]]:
        return [cls]

    @classmethod
    @time_func
    def cast(cls, items: list[Self]):
        for item in items:
            if isinstance(item, cls):
                return item

    @property
    def type_id(self):
        return self.__class__

    @time_func
    def get[T:Component](self, ctype: Type[T]) -> T | None:
        return self.db.get_table(ctype).read(self.eid)



import builtins
import dataclasses
from typing import Type, TypeVar

from easy_ecs_sim.component import Component
from easy_ecs_sim.signature import Signature

ComponentSet = Component | list[Component] | Signature | None

T = TypeVar('T')


def flatten_components(item: ComponentSet):
    if isinstance(item, Component):
        return [item]
    if isinstance(item, Signature):
        return item.to_components()
    return item


def column_mapping[T: dataclasses.dataclass](ctype: Type[T], prefix: str = ''):
    if not prefix:
        prefix = f'{ctype.__name__}'
    res = {}

    for field in dataclasses.fields(ctype):
        subtype = field.type
        name = f'{prefix}.{field.name}'

        if dataclasses.is_dataclass(field.type):
            res.update(column_mapping(subtype, prefix=name))
        else:
            match field.type:
                case builtins.int:
                    res[name] = field.type
                case builtins.float:
                    res[name] = field.type
                case builtins.bool:
                    res[name] = field.type
                case _:
                    res[name] = field.type
    return res



from functools import cache
from typing import Type, override, TypeVar

from easy_kit.timing import time_func

from easy_ecs_sim.component import Component
from easy_ecs_sim.signature import Signature
from easy_ecs_sim.storage.database import Database
from easy_ecs_sim.storage.demography import Demography
from easy_ecs_sim.storage.id_generator import IdGenerator
from easy_ecs_sim.storage.index import Index
from easy_ecs_sim.types import EntityId
from easy_ecs_sim.utils import ComponentSet, flatten_components

EID_GEN = IdGenerator()

T = TypeVar('T')


class MyDatabase(Database):
    def __init__(self):
        self._entities: set[EntityId] = set()
        self.tables: dict[Type[Component | Signature], Index] = {}
        self.dirty: Demography = Demography()

    @override
    @time_func
    def find_any[T: Component](self, what: Type[T], having: list[Type[Component]] = None) -> T:
        table = self.get_table(what)
        eids = table.entities
        for _ in having or []:
            eids.intersection_update(self.get_table(_).entities)

        if eids:
            eid = next(iter(eids))
            return table.read(eid)

    @override
    def entities(self):
        return self._entities

    @override
    def create_all(self, items: list[ComponentSet]):
        items = list(map(flatten_components, items))
        for components in items:
            eid = EID_GEN.new_id()
            for _ in components:
                _.eid = eid

        self.dirty.with_birth(items)

    @override
    def destroy_all(self, items: Component | Signature | list[Component | Signature]):
        self.dirty.with_death(items)

    @override
    @cache
    def get_table[T:Component | Signature](self, ttype: Type[T]) -> Index[T]:
        if ttype not in self.tables:
            self.tables[ttype] = Index(ttype=ttype)
        return self.tables[ttype]

    @time_func
    def union_entities(self, signature: list[Type[Component]]):
        res = set()
        for _ in signature:
            res.update(self.tables[_].by_entity.keys())
        return res

    @time_func
    def intersect_entities(self, signature: list[Type[Component]]):
        if not signature:
            return set()
        first = signature[0]
        if first not in self.tables:
            return set()

        res = self.tables[first].entities
        for _ in signature[1:]:
            res.intersection_update(self.tables[_].entities)
            if not res:
                return res
        return res

    # -----------------------------------------------------------------------------

    @time_func
    def update_demography(self, status: Demography):
        death = status.death

        self._entities.difference_update(death)
        for table in self.tables.values():
            table.destroy_all(death)

        birth = filter(None, status.birth)
        for components in birth:
            self._entities.add(components[0].eid)
            for c in components:
                if c is None:
                    continue
                c.db = self
                ctype = c.type_id
                self.get_table(ctype).create(c)



import functools
from typing import Type

from easy_kit.timing import time_func
from pydantic import model_validator

from easy_kit.my_model import MyModel
from easy_ecs_sim.component import Component
from easy_ecs_sim.types import EntityId


class Signature(MyModel):
    @model_validator(mode='after')
    def post_init(self):
        for name, _ in self.model_fields.items():
            if not isinstance(_.annotation, type(Component)):
                raise ValueError(f'Signature: {name} type is [{_.annotation}] (must be a Component)')
        return self

    @property
    def eid(self):
        name = self.field_names()[0]
        item = getattr(self, name)
        return EntityId(item.eid)

    def to_components(self) -> list[Component]:
        return list(filter(None, [
            getattr(self, name)
            for name in self.field_names()
        ]))

    @classmethod
    @time_func
    @functools.lru_cache()
    def signature(cls) -> list[Type[Component]]:
        return [cls.model_fields[_].annotation for _ in cls.field_names()]

    @classmethod
    @functools.lru_cache()
    def field_names(cls):
        return list(sorted(cls.model_fields.keys()))

    @classmethod
    @time_func
    # @functools.lru_cache()
    def field_mapping(cls):
        return dict(zip(cls.signature(), cls.field_names()))

    @classmethod
    def match(cls, items: list[Component]):
        mapping = cls.field_mapping()
        for item in items:
            name = mapping.pop(item.type_id)
        return len(mapping) == 0

    @classmethod
    @time_func
    def cast(cls, items: list[Component]):
        mapping = cls.field_mapping()
        if len(items) < len(mapping):
            return None

        res = cls.model_construct()
        for item in items:
            name = mapping.pop(item.type_id, None)
            if name is not None:
                setattr(res, name, item)
        if len(mapping) > 0:
            return None
        return res



from abc import abstractmethod
from dataclasses import field
from typing import Type, override, TypeVar

from easy_kit.my_model import MyModel
from pydantic import model_validator

from easy_ecs_sim.component import Component
from easy_ecs_sim.context import Context
from easy_ecs_sim.signature import Signature
from easy_ecs_sim.storage.database import Database
from easy_ecs_sim.timebox_manager import TimeboxManager

T = TypeVar('T')


class BaseSystem(MyModel):
    periodicity_sec: float = 0  # expected (or minimum) time between two updates

    def at_interval(self, periodicity_sec: float):
        self.periodicity_sec = periodicity_sec
        return self

    @abstractmethod
    def update(self, ctx: Context, db: Database, dt: float):
        ...


class System[T: Signature | Component](BaseSystem):
    _signature: Type[T] = None
    max_time_per_step: float | None = None
    timebox: TimeboxManager[T] | None = None

    @model_validator(mode='after')
    def post_init(self):
        if self.max_time_per_step is not None:
            self.timebox = TimeboxManager(
                signature=self._signature,
                max_time_per_step=self.max_time_per_step
            )
        return self

    @property
    def sys_id(self):
        return self.__class__.__name__

    def update_single(self, ctx, db: Database, item: T, dt: float):
        pass

    def register(self, ctx: Context, item: T):
        pass

    def unregister(self, ctx: Context, item: T):
        pass

    @override
    def update(self, ctx: Context, db: Database, dt: float):
        if self.timebox:
            dt_fixed = dt + self.timebox.dt_bonus
            for item in self.timebox.iter(db, dt):
                self.update_single(ctx, db, item, dt_fixed)
        else:
            for item in db.get_table(self._signature).iter():
                self.update_single(ctx, db, item, dt)


class SystemBag(System):
    name: str
    steps: list[System] = field(default_factory=list)

    def update(self, ctx: Context, db: Database, dt: float):
        for _ in self.steps:
            _.update(ctx, db, dt)



from abc import ABC, abstractmethod
from typing import Type, TypeVar

from easy_ecs_sim.component import Component
from easy_ecs_sim.signature import Signature
from easy_ecs_sim.storage.index import Index
from easy_ecs_sim.types import EntityId
from easy_ecs_sim.utils import ComponentSet

T = TypeVar('T')


class Database(ABC):

    @abstractmethod
    def entities(self) -> set[EntityId]:
        ...

    @abstractmethod
    def create_all(self, items: list[ComponentSet]):
        ...

    @abstractmethod
    def destroy_all(self, items: Component | Signature | list[Component | Signature]):
        ...

    @abstractmethod
    def get_table[T:Component | Signature](self, ttype: Type[T]) -> Index:
        ...

    @abstractmethod
    def find_any[T: Component](self, what: Type[T], having: list[Type[Component]]) -> T:
        ...
